/* ----------------------------------------------------------------------------
 * mask.C
 */

extern "C" {
#include <stdio.h>
#include <string.h>
#include <memory.h>
#ifndef MINT
#include <malloc.h>
#endif

extern int atoi(char *t);
}

#include "twin.h"

/* ------------------------------------------------------------------------- */

ViewField::ViewField (char *InitName,
		void (*InitCheck)(ViewField *v,int ch),
		Field *InitField,
		int InitX, int InitY, int InitB, int InitH)
	: View (InitName, InitX,InitY, InitB,InitH)
{	check  = InitCheck;
	if (h != 1) h=1;
	visible = 1;
	editbuff=0;
	editflag=0;
	field = InitField;
	}

void ViewField::expose(int ExX,int ExY,int ExB,int ExH)
{	char *t;

        if (view_debug_expose)
                fprintf(stdout,"%sFieldExpose %s(%d,%d,%d,%d)\n\r",
                        visible ? "" : "No", name,ExX,ExY,ExB,ExH);

	if (ExY != 0) return;
	
	if (editbuff) t=strdup(editbuff);
	else {	t=(char*)setmalloc(b+1);
		field->unpack(t);
		}

	if ((int)strlen(t) > ExX+ExB) t[ExX+ExB] = 0;
	
	if (VRoot.focus == this) {
		tw_term_set(RealX(ExX),RealY(ExY),TW_UNDER,ExB);
		tw_term_puts(RealX(ExX),RealY(ExY),TW_UNDER,t+ExX);
		}
	else	tw_term_puts(RealX(ExX),RealY(ExY), TW_NORMAL,t+ExX);

	free(t);
	}

void ViewField::event(int EvX,int EvY,int EvKey)
{	int i;

        if (view_debug_event) 
                fprintf(stdout,"%sFieldEvent %s(%d,%d,%d)\n\r",
                        visible ? "" : "No", name,EvX,EvY,EvKey);

	if (visible != 1) return;

	if (!editbuff) {
		editbuff=(char*)setmalloc(b+1);
		editflag = 1;
		field->unpack(editbuff);
		}

	if (EvX<0) EvX=0;
	if (EvY<0) EvY=0;
	if (EvX>=b) EvX=b-1;
	if (EvY>=h) EvY=h-1;

	focus_in(EvKey);

	if ((EvKey >=32) && (EvKey <= 255)) {
		if (editflag && !EvX) {
			memset(editbuff,0,b+1);
			editbuff[0]=EvKey;
			EvX=0;
			}
		else {	if ((int)strlen(editbuff)<EvX) EvX = strlen(editbuff);
			for (i=b; i>EvX; i--) editbuff[i] = editbuff[i-1];
			editbuff[b] = 0; editbuff[EvX]=EvKey;
			}
		editflag = 0;

		field->pack(editbuff);

		VRoot.expose(RealX(0),RealY(0),b,h);
		if (EvX<b-1) EvX++;
		}
	else switch (EvKey) {
		case TERM_LE : if (EvX>0)	EvX--; break;
		case TERM_RE : if (EvX<b-1)	EvX++; break;
		case TERM_BS : if (EvX>0)	EvX--;
		case TERM_DEL: 
			editflag = 0;
			for (i=EvX;i<b;i++) editbuff[i] = editbuff[i+1];
			editbuff[b] = 0;
			field->pack(editbuff);
			VRoot.expose(RealX(0),RealY(0),b,h);
			break;
		}

	set_cursor(EvX,EvY);
	}

void ViewField::focus_in(int reason)
{	if (VRoot.focus != this) {
		View::focus_in(reason);
		VRoot.expose(RealX(0),RealY(0),b,h);
		}
	set_cursor(0,0);
	}

void ViewField::focus_out(int reason)
{	if (editbuff) {
		field->pack(editbuff);
		free(editbuff);
		editbuff=0;
		}
	VRoot.focus = &VRoot;
	VRoot.expose(RealX(0),RealY(0),b,h);
	if (check) (*(check))((ViewField *)this,reason);
	}

/* -------------------------------------------------------------------------- */

ViewMask::ViewMask(char *InitName, char **InitBackground, MaskLoad *InitLoad,
		int InitX, int InitY,
		int InitB, int InitH)
        : ViewBack(InitName,InitX,InitY,InitB,InitH)
{	background = InitBackground;
	focus = 0;
	viewlist=0;
	if (InitLoad) load(InitLoad);
	}

void ViewMask::load(MaskLoad *load)
{	struct ViewList **vl;
	char **back;
	int by;
	char *p;
	Field *f;
	int mx;

	mx = 0;

	if (viewlist) return;

	vl = &viewlist;
	while (load->name != 0) {
		f = (*(load->konstr)) (load->name, load->buff,load->len);
		*vl = (struct ViewList *)setmalloc(sizeof(struct ViewList));
		(*vl)->view = new ViewField(load->name, load->check, f,
			0,0, load->len, 1);
		(*vl)->view->transaction = load->trans;
		(*vl)->next = 0;

		by=0; back=background;

		while (back && *back) {
			p = strchr(*back,'$');
			if (p) {
				*p = ' ';
				(*vl)->view->y = by;
				(*vl)->view->x = (p-*back);
				(*vl)->view->map(this);
				(*vl)->view->visible = 1;
				back=0;
				if (mx < (*vl)->view->x+(*vl)->view->b+1)
					mx = (*vl)->view->x+(*vl)->view->b+1;
				}
			else {	back++; by++; }
			}
		if (back) {
		    by=0; back=background;
		    while (back && *back) {
			p = strchr(*back,'#');
			if (p) {
				*p = ' ';
				(*vl)->view->y = by;
				(*vl)->view->x = (p-*back);
				(*vl)->view->map(this);
				(*vl)->view->visible = 2;
				back=0;
				if (mx < (*vl)->view->x+(*vl)->view->b+1)
					mx = (*vl)->view->x+(*vl)->view->b+1;
				}
			else {	back++; by++; }
			}
		    }
		
		load++;
		vl = &((*vl)->next);
		}
	if (background && (!b || !h)) {
		b= mx;

		by=0; back=background;
		while (back && *back) {
			back++; by++;
			}
		h = by;
		}
	}

void ViewMask::event(int EvX,int EvY,int EvKey)
{       struct ViewList *v;
        struct View *tmp;

        if (view_debug_event) 
                fprintf(stdout,"%sMaskEvent %s(%d,%d,%d)\n\r",
                        visible ? "" : "No", name,EvX,EvY,EvKey);

	focus_in(0);

	if (viewlist) {
		if (!focus) focus = viewlist->view;
		v = viewlist; tmp = 0;

		if ((EvKey == TERM_DN) || (EvKey == TERM_CR) ||
		    (EvKey == TERM_UP)) {
			if (VRoot.focus == focus) {
				focus->focus_out(EvKey);
				if (VRoot.focus != &VRoot)
					EvKey = 0;
				}
			}

		if (EvKey == TERM_UP) {
			View *f = focus;


			if (f == focus)
                	    while (v) {
                       		if (v->next &&
				   (v->next->view == focus)) {
                               		tmp = v->view;
					break;
					}
                       		v = v->next;
                               	}
			if (tmp && (tmp->visible == 1))
				focus = tmp;
			else	focus = viewlist->view;

			focus->focus_in(EvKey);
			}
		else
		if ((EvKey == TERM_DN) || (EvKey == TERM_CR)) {
			View *f = focus;

			if (VRoot.focus == focus)
				focus->focus_out(EvKey);

			if (f == focus)
                	    while (v) {
				if (v->view == focus) {
					if (v->next) 
                               			tmp = v->next->view;
					break;
					}
                        	v = v->next;
                               	}
			if (tmp && (tmp->visible == 1))
				focus = tmp;
			else	focus = viewlist->view;

			focus->focus_in(EvKey);
			}
		else
		if (EvKey) {
        		if ((tmp=find(EvX,EvY)) && (tmp->visible == 1)) {
				focus = tmp;
				focus->focus_in(EvKey);
				focus->event(EvX-focus->x,EvY-focus->y,EvKey);
				}
			else {	focus = viewlist->view;
				focus->focus_in(EvKey);
				focus->event(EvX-focus->x,EvY-focus->y,EvKey);
				}
			}
		}
        }

void ViewMask::focus_in(int reason)
{	if (viewlist && (VRoot.focus->parent != this)) {
                focus = viewlist->view;
                viewlist->view->focus_in(reason);
                }
	}

