/* ----------------------------------------------------------------------------
 * tw_types.h (c) 92 Michael Kraehe
 */

#ifndef TWTYPES
#define TWTYPES

extern "C" {
	#include "fcntl.h"
	}

extern class View;

/* Field types ------------------------------------------------------------- */

class Field {
	public :
		char *name;
		int len;
                Field(char *InitName, int InitLen);

		virtual void pack(char *str);
		virtual void unpack(char *str);
		};

class String : public Field {
	public :
                char *buff;

                String(char *InitName, char *InitStr,int InitLen);

		virtual void pack(char *str);
		virtual void unpack(char *str);
		};

class Short : public Field {
	public :
                short *buff;

                Short(char *InitName, short *InitStr,int InitLen);

		virtual void pack(char *str);
		virtual void unpack(char *str);
		};

class Long : public Field {
	public :
                long *buff;

                Long(char *InitName, long *InitStr,int InitLen);

		virtual void pack(char *str);
		virtual void unpack(char *str);
		};

class Calc : public Field {
	public :
                long *buff;

                Calc(char *InitName, long *InitStr,int InitLen);

		virtual void pack(char *str);
		virtual void unpack(char *str);
		};

class Date : public Field {
	public :
                long *buff;

                Date(char *InitName, long *InitStr,int InitLen);

		virtual void pack(char *str);
		virtual void unpack(char *str);
		};

extern Field *NewString(char *Name, char *Buff, int Len);
extern Field *NewShort(char *Name, char *Buff, int Len);
extern Field *NewLong(char *Name, char *Buff, int Len);
extern Field *NewCalc(char *Name, char *Buff, int Len);
extern Field *NewDate(char *Name, char *Buff, int Len);

/* ------------------------------------------------------------------------- */

class List {
	public :
		char **lines;
		int nlines;
		int inc;
		int freeflag;
	
		List(int InitInc);
		virtual ~List();

		virtual void add_pos(char *line);
		virtual void add_pos(int pos,char *line);
		virtual void del_pos(int pos);

		virtual void add_line(char *line);
		virtual void add_line(int pos,char *line);

		virtual void purge(int freefl = -1);
	};

class Text : public List {
	public :
		Text(int InitInc);
		virtual ~Text();
	};

struct HashLinks {
	char *key;
	int next;
	char *val;
	};

class HashList : public List {
	public :
		int *hashtab;
		int hashmod;

		HashList(int InitInc, int InitHash = 37);
		virtual ~HashList();

		int Hash(char *key);
		void add_buck(char *key,char *val);
		int ind_buck(char *key,char *k2=0);
		char *get_buck(char *key,char *k2=0);
		char *get_key(int n);
		char *get_val(int n);
	};

/* -------------------------------------------------------------------------- */

class LineStream {
	public :
		int pos;
		char *StreamName;
		virtual int readline(char *line,int len);
		virtual int writeline(char *line);
	};

class FileStream : public LineStream {
	public :
		int fd;
		int flags;
		int closeflag;
		char buff[512];
		char *bp;
		int bl;
		int maxread;

		FileStream(int InitFd,int InitFlags = O_RDONLY);
		FileStream(char *FileName,int InitFlags = O_RDONLY);

		virtual int readline(char *line,int len);
		virtual int writeline(char *line);

		virtual ~FileStream();
	};

class TextStream : public LineStream {
	public :
		Text *text;
		int flags;

		TextStream(Text *InitText,int InitFlags = O_RDONLY);
		virtual ~TextStream();

		virtual int readline(char *line,int len);
		virtual int writeline(char *line);
	};

/* Aggrgate type ----------------------------------------------------------- */

struct RecLoad {
	char *name;
	char *buff;
	int  len;
	Field *((*konstr)(char *Name, char *Buff, int Len));
	};

struct RecDesc {
	char *name;
	char *buff;
	int  len;
	Field *field;
	Field *((*konstr)(char *Name, char *Buff, int Len));
	};

class Rec {
	public:
		char *name;
		char *buff;
		int len;
		struct RecDesc *rec;
		int nfields;
		Text *last_transaction;

		Rec(char *InitName, char *Buff, int Len,
			struct RecLoad *InitLoad);
		virtual void load(struct RecLoad *InitLoad);

		virtual void export (LineStream *Stream,List *fields =0);
		virtual void export (char *line,List *fields =0);
		virtual int import (LineStream *Stream,List *fields =0);
		virtual int import (char *line,List *fields =0);
		int compare(char **transaction);
	};

class Cursor : public Rec {
	public:
		int cursor;
		int nrecs;
		int seek;

		Cursor(char *InitName, char *Buff, int Len,
			struct RecLoad *InitLoad);
		virtual ~Cursor();

		virtual void clean();
		virtual void flush();
		virtual void block(int r);

		virtual void add();
		virtual void first();
		virtual void last();
		virtual void prev();
		virtual void next();

		virtual void export (LineStream *Stream,List *fields =0);
		virtual int import (LineStream *Stream,List *fields =0);
	};

extern int makepipe(int *in,int *out,int *err);

#endif
