/* ----------------------------------------------------------------------------
 * ViewField.C
 */

int rec_import_test =0;

extern "C" {
	#include <string.h>
	#include <memory.h>
	#ifndef MINT
	#include <malloc.h>
	#endif
	#include <stdio.h>
	#include <ctype.h>
	#include <unistd.h>
	#include <fcntl.h>
	#include <signal.h>

	#ifndef O_BINARY
	#define O_BINARY 0
	#endif

	#include <assert.h>

	#include <time.h>
	}

#include "twin.h"

extern "C" {
	extern int atoi(char *t);
	extern long atol(char *t);
	}

/* -------------------------------------------------------------------------- */

Field::Field (char *InitName, int InitLen)
{	name=strdup(InitName);
	len=InitLen;
	}

void Field::pack(char *str) 
{	memset(str,' ',len); str[len]=0;
	}

void Field::unpack(char *str)
{	}

/* -------------------------------------------------------------------------- */

String::String(char *InitName, char *InitStr, int InitLen)
	: Field(InitName,InitLen)
{	buff = InitStr;
	}

void String::unpack(char *str)
{	memset(str,0,len+1);
	strncpy(str,buff,len);
	}

void String::pack(char *str)
{	memset(buff,0,len);
	strncpy(buff,str,len);
	}

Field* NewString(char *Name, char *Buff, int Len)
{	String *v;

	v = new String(Name,Buff,Len);
	return((Field *)v);
	}

/* -------------------------------------------------------------------------- */

Short::Short(char *InitName, short *InitShort, int InitLen)
	: Field(InitName,InitLen)
{	buff = InitShort;
	}

void Short::unpack(char *str)
{	memset(str,0,len+1);
	/* if (*buff) */ sprintf(str,"%*d",len,*buff);
	}

void Short::pack(char *str)
{	char *p;
	p=str; while (*p==' ') p++;
	*buff=atoi(p);
	}

Field* NewShort(char *Name, char *Buff, int Len)
{	Short *v;

	v = new Short(Name,(short *)Buff,Len);
	return((Field *)v);
	}

/* -------------------------------------------------------------------------- */

Long::Long(char *InitName, long *InitLong, int InitLen)
	: Field(InitName,InitLen)
{	buff = InitLong;
	}

void Long::unpack(char *str)
{	memset(str,0,len+1);
	/* if (*buff) */ sprintf(str,"%*ld",len,*buff);
	}

void Long::pack(char *str)
{	char *p;
	p=str; while (*p==' ') p++;
	*buff=atol(p);
	}

Field* NewLong(char *Name, char *Buff, int Len)
{	Long *v;

	v = new Long(Name,(long *)Buff,Len);
	return((Field *)v);
	}

/* -------------------------------------------------------------------------- */

Calc::Calc(char *InitName, long *InitCalc, int InitLen)
	: Field(InitName,InitLen)
{	buff = InitCalc;
	}

void Calc::unpack(char *str)
{	memset(str,0,len+1);
	/* if (*buff) */
		sprintf(str,"%*d.%2.2d",len-3,*buff/100l,*buff%100l);
	}

void Calc::pack(char *str)
{	char *p;
	p=str; while (*p==' ') p++;
	*buff=atol(p)*100l;
	if (p=strchr(str,'.')) {
		if (isdigit(p[1])) *buff += (p[1]-'0') * 10l;
		if (isdigit(p[2])) *buff += (p[2]-'0');
		}
	}

Field* NewCalc(char *Name, char *Buff, int Len)
{	Calc *v;

	v = new Calc(Name,(long *)Buff,Len);
	return((Field *)v);
	}

/* -------------------------------------------------------------------------- */

Date::Date(char *InitName, long *InitDate, int InitLen)
	: Field(InitName,InitLen)
{	buff = InitDate;
	}

void Date::unpack(char *str)
{	memset(str,0,len+1);
	if (*buff) {
		if (len == 5)
			sprintf(str,"%2.2d.%2.2d",
				*buff%100l,
				(*buff%10000l)/100l
				);
		if (len == 8)
			sprintf(str,"%2.2d.%2.2d.%2.2d",
				*buff%100l,
				(*buff%10000l)/100l,
				*buff/10000l
				);
		if (len == 10)
			sprintf(str,"%2.2d.%2.2d.%4.4d",
				*buff%100l,
				(*buff%10000l)/100l,
				*buff/10000l
				);
		}
	}

void Date::pack(char *str)
{	char *p,*q;
	char date[11];
	long day,mon,year;
	struct tm *tm_buf;
	long clock;

	clock = time((long*)0);
	tm_buf = localtime(&clock);

	day = tm_buf->tm_mday;
	mon = tm_buf->tm_mon+1;
	year= tm_buf->tm_year+1900;

	if (p=strchr(str,'.')) {
		day = atol(str);
		q = p+1;
		mon = atol(q);
		if (p=strchr(q,'.')) {
			year = atol(p+1);
			if (year<100) year+=1900l;
			}
		}
	else {	clock = atol(str);
		if (clock > 1000000l) {
			year = clock % 10000l;
			mon = (clock % 1000000l)	/ 10000l;
			day = (clock % 100000000l)	/ 1000000l;
			}
		else
		if (clock > 10000l) {
			year = clock % 100l + 1900;
			mon = (clock % 10000l)		/ 100l;
			day = (clock % 1000000l)	/ 10000l;
			}
		else
		if (clock > 100l) {
			mon =  clock % 100l;
			day =  clock / 100l;
			}
		else
		if (clock > 0l) {
			day =  clock % 100l;
			}
		else {	day = mon = year = 0l;
			}
		}

	if (buff)
		*buff = year*10000l + mon*100l + day;
	}

Field* NewDate(char *Name, char *Buff, int Len)
{	Date *v;

	v = new Date(Name,(long *)Buff,Len);
	return((Field *)v);
	}

/* -------------------------------------------------------------------------- */

Rec::Rec(char *InitName,char *InitBuff,int InitLen,struct RecLoad *InitLoad)
{	name = InitName;
	buff = InitBuff;
	len  = InitLen;
	last_transaction=new Text(1);
	if (InitLoad)
		load(InitLoad);
	else {	rec=0;
		nfields=0;
		}
	}

void Rec::load(struct RecLoad *InitLoad)
{	int n;
	struct RecLoad *load;

	load = InitLoad; n=0;
	while (load -> name != 0) {
		load ++;
		n++;
		}

	rec = (struct RecDesc *) setmalloc(sizeof(struct RecDesc)*(n+1));
	
	load = InitLoad; n=0;
	while (load -> name != 0) {
		/* fprintf(stderr,"Rec::Load(%s)\n",load->name); */
		rec[n].name   = load->name;
		rec[n].buff   = load->buff;
		rec[n].len    = load->len;
		rec[n].konstr = load->konstr;
		rec[n].field  = (*(load->konstr)) (load->name, load->buff,
				load->len);
		load ++;
		n++;
		}
	nfields = n;
	}

void Rec::export (char *line,List *fields)
{	int i;
	char str[256];

	*line = 0;

	if (fields)
	    for (i=0; i<fields->nlines; i++) {
		((Field*)(fields->lines[i]))->unpack(str);
		strcat(line,str);
		if (i+1<fields->nlines) strcat(line,"\t");
		}
	else
	    for (i=0; i<nfields; i++) {
	    	rec[i].field->unpack(str);
		strcat(line,str);
		if (i+1<fields->nlines) strcat(line,"\t");
		}
	}


int Rec::import (char *line,List *fields)
{	int i;
	char str[256];
	char *p,*q;
	int strl;

	i=0; p=line;
	memset(str,0,256); q=str; strl = 0;

	while (*p) {
		if (*p == '\t') {
			if (fields) {
				if (i < fields->nlines)
				   ((Field*)(fields->lines[i]))->pack(str);
				}
			else {	if (i<nfields)
					rec[i].field->pack(str);
				}

			i++; memset(str,0,256); q=str; strl = 0;
			}
		else
		if ((*p == '\n') || (*p == '\r'))
			break;
		else {	if (strl<254) {
				*q++ = *p;
				strl++;
				}
			}

		p++;
		}

	if (fields) {
		if (i < fields->nlines)
		   ((Field*)(fields->lines[i]))->pack(str);
		}
	else {  if (i<nfields)
			rec[i].field->pack(str);
		}

	return(1);
	}

void Rec::export (LineStream *stream,List *fields)
{	char line[1024];

	export (line,fields);
	stream->writeline(line);
	}

int Rec::import (LineStream *stream,List *fields)
{	char line[1024];
	int ret;

	ret = stream->readline(line,1023);

	if (ret >=0) import (line,fields);

	return(ret);
	}

int Rec::compare(char **transaction)
{	char **p,**q;

	p=transaction; q=last_transaction->lines;
	while (*p && *q && !strcmp(*p,*q)) {
		/*
		tw_term_error(3,"%s",*p);
		tw_term_error(4,"%s",*p);
		*/
		p++; q++;
		}
	if (!*p && !*q) return(0);
	
	delete last_transaction;
	last_transaction= new Text(p-transaction);

	p=transaction;
	while (*p) {
		last_transaction->add_line(*p);
		p++;
		}
	return(1);
	}
	

/* ------------------------------------------------------------------------- */

Cursor::Cursor(char *InitName,char *InitBuff,int InitLen,
		struct RecLoad *InitLoad)
	:Rec(InitName,InitBuff,InitLen,InitLoad)
{	cursor=0;

	clean();
	}

Cursor::~Cursor()
{	if (cursor>0) close(cursor);
	}

void Cursor::clean()
{	char fname[64];

	if (cursor>0) close(cursor);

	sprintf(fname,"/tmp/%d.%s.cur",getpid(),name);
	cursor = open(fname,O_RDWR | O_CREAT | O_TRUNC | O_BINARY,0666);
	unlink(fname);

	nrecs=0;
	seek=0;

	if (buff) memset(buff,0,len);
	}

void Cursor::flush()
{	if (seek < nrecs) {
		lseek(cursor,(long)(seek)*(long)(len),0);
		write(cursor,buff,len);
		}
	}

void Cursor::block(int r)
{	flush();
	if (r < nrecs) {
		seek=r;
		lseek(cursor,(long)(seek)*(long)(len),0);
		read(cursor,buff,len);
		}
	else	memset(buff,0,len);
	seek = r;
	}

void Cursor::add()
{	if (seek == nrecs) {
		nrecs++;
		flush();
		}
	seek=nrecs; memset(buff,0,len);
	}

void Cursor::first()
{	block(0);
	}
	
void Cursor::last()
{	if (nrecs)
		block(nrecs-1);
	else	block(0);
	}

void Cursor::next()
{	if (nrecs > seek+1)
		block(seek+1);
	else	last();
	}

void Cursor::prev()
{	if (seek>0) block(seek-1);
	else	first();
	}

void Cursor::export(LineStream *stream,List *fields)
{	seek=0;
	while (seek < nrecs) {
		block(seek);
		Rec::export(stream,fields);
		seek++;
		}
	}

int Cursor::import(LineStream *stream,List *fields)
{	assert(this!=0);

	clean();

	/* add(); */
	while (Rec::import(stream,fields)>=0) add();
	/* nrecs--; */ first();

	return(nrecs);
	}

/* ------------------------------------------------------------------------- */

List::List(int InitInc)
{	inc = InitInc ? InitInc : 16;

	lines = (char **)setmalloc(inc * sizeof(char *));
	nlines = 0;

	assert(lines[nlines]==0);
	freeflag = 0;
	}

List::~List()
{	if (freeflag) purge();
	free(lines);
	}

void List::add_pos(char *line)
{	char **l;

	if (nlines+1 >= inc) {
		l = (char **)setmalloc(2 * inc * sizeof(char *));
		memcpy(l,lines,inc * sizeof(char *));
		free(lines); lines = l;
		inc = inc * 2;
		}
	if (line) lines[nlines]=line;
	nlines++;

	assert(lines[nlines]==0);
	assert(nlines<inc);
	}

void List::add_pos(int pos,char *line)
{	char **l;
	int i;

	if (nlines+1 >= inc) {
		l = (char **)setmalloc(2 * inc * sizeof(char *));
		memset(l,0,2 * inc * sizeof(char *));
		memcpy(l,lines,inc * sizeof(char *));
		free(lines); lines = l;
		inc = inc * 2;
		}
	for (i = nlines -1; i>=pos; i--)
		lines[i+1] = lines[i];
	if (line) lines[pos]=line;
	nlines++;

	assert(lines[nlines]==0);
	assert(nlines<inc);
	}

void List::del_pos(int pos)
{	int i;

	if (lines[pos] && freeflag) free(lines[pos]);

	for (i = pos; i < nlines; i++) lines[i] = lines[i+1];
	nlines--;
	lines[nlines]=0;

	assert(lines[nlines]==0);
	}

void List::purge(int freefl)
{	int i;

	for (i=0; i<nlines;i++) 
		if (lines[i]) {
			if (freefl < 0 ? freeflag : freefl)
				free(lines[i]);
			lines[i]=0;
			}
	lines[0]=0;
	nlines=0;

	assert(lines[nlines]==0);
	}

void List::add_line(char *line)
{	add_pos(strdup(line));
	}

void List::add_line(int pos,char *line)
{	add_pos(pos,strdup(line));
	}

/* ------------------------------------------------------------------------- */

Text::Text(int InitInc) : List(InitInc)
{	freeflag = 1;
	}

Text::~Text()
{	}

/* ------------------------------------------------------------------------- */

HashList::HashList(int InitInc,int InitHash) : List(InitInc)
{	int i;

	hashmod = InitHash;
	hashtab = (int *)setmalloc(hashmod * sizeof(int));
	for (i=0;i<hashmod;i++) hashtab[i] = -1;

	freeflag=1;
	}

HashList::~HashList()
{	free(hashtab);
	}

int HashList::Hash(char *key)
{	char *p;
	int h;

	h=0; p=key;

	while (*p) {
		h = (h + *p) % hashmod;
		p++;
		}

	return (h);
	}

void HashList::add_buck(char *key,char *val)
{	struct HashLinks *hl;
	int h;

	h = Hash(key);

	hl = (struct HashLinks *)setmalloc(sizeof(struct HashLinks));
	hl->key = key;
	hl->val = val;
	hl->next = hashtab[h];

	hashtab[h] = nlines;
	add_pos((char*)hl);
	}

int HashList::ind_buck(char *key,char *k1)
{	struct HashLinks *hl;
	int h;
	char *p;

	if (k1) {
		p = setmalloc(strlen(key)+strlen(k1)+2);
		sprintf(p,"%s.%s",key,k1);
		}
	else	p = key;

	/*
	fprintf(stderr,"ind_buck(%s)\n",p);
	*/

	h = hashtab[Hash(p)];

	while (h>=0) {
		hl = (struct HashLinks *)lines[h];
		if (!strcmp(p,hl->key)) break;
		h = hl->next;
		}

	if (k1) free(p);

	/*
	fprintf(stderr,"ind_buck(%d)\n",h);
	*/

	return(h);
	}

char *HashList::get_buck(char *key,char *k1)
{	struct HashLinks *hl;
	int h;
	
	if ((h = ind_buck(key,k1))<0) return(0);

	hl = (struct HashLinks *)lines[h];

	/*
	fprintf(stderr,"get_buck(%d,%d,%s)\n",hl,hl->val,hl->val);
	*/
	return(hl->val);
	}

char *HashList::get_key(int n)
{	struct HashLinks *hl;

	if (n>=nlines) return(0);
	hl = (struct HashLinks *)lines[n];
	if (!hl) return(0);
	return(hl->key);
	}

char *HashList::get_val(int n)
{	struct HashLinks *hl;

	if (n>=nlines) return(0);
	hl = (struct HashLinks *)lines[n];
	if (!hl) return(0);
	return(hl->val);
	}

/* ------------------------------------------------------------------------- */

int LineStream::readline(char *line,int len)
{	return(-1);
	}

int LineStream::writeline(char *line)
{	return(-1);
	}

#define READBUFF 512

FileStream::FileStream(int InitFd,int InitFlags)
{	char str[64];

	sprintf(str,"File-ID=%d", InitFd);
	StreamName = strdup(str);

	fd = InitFd;
	flags = InitFlags;
	closeflag = 0;
	pos = 0;
	bp=buff;
	bl=0;
	maxread=READBUFF;
	memset(buff,0,512);
	}

FileStream::FileStream(char *FileName,int InitFlags)
{	fd = open(FileName,InitFlags,0666);
	flags = InitFlags;
	closeflag = 1;
	pos = 0;
	bp=buff;
	bl=0;
	maxread=READBUFF;
	memset(buff,0,512);

	StreamName = strdup(FileName);
	}

FileStream::~FileStream()
{	if (closeflag) close(fd);
	free(StreamName);
	}

int FileStream::readline(char *line,int len)
{       char *p;

	if (((flags&O_ACCMODE) != O_RDWR) && ((flags&O_ACCMODE) != O_RDONLY))
		return(-1);

	if (bp-buff >= bl) {
		bl = read(fd,buff,maxread);
		bp = buff;
		}

	memset(line,0,len);

	if (bl<=0) return(-1);

        p=line;
        while(p < line+len) {
		if (bp-buff >= bl) {
			bl = read(fd,buff,maxread);
			bp = buff;
			if (bl<=0) break;
			}
                if ((*bp != '\r') && (*bp != '\n')) *p++=*bp;
                if (*bp++ == '\n') break;
                }

        *p=0; pos++;

        return(p-line);
        }

int FileStream::writeline(char *line)
{	int ret;

	if (((flags&O_ACCMODE) != O_RDWR) && ((flags&O_ACCMODE) != O_WRONLY))
		return(-1);

	signal(SIGPIPE,SIG_IGN);

	if (line) {
		ret=write(fd,line,strlen(line));
		}
	else	ret=0;

	write(fd,"\n",1);
	ret++;

	if (ret>0) pos++;

	return(ret);
	}

TextStream::TextStream(Text *InitText,int InitFlags)
{	text = InitText;
	pos = 0;
	flags = InitFlags;
	StreamName=strdup("File-Id:Text");
	}

TextStream::~TextStream()
{	free(StreamName);
	}

int TextStream::readline(char *line,int len)
{	int ret = 0;

	if (((flags&O_ACCMODE) != O_RDWR) && ((flags&O_ACCMODE) != O_RDONLY))
		return(-1);

	memset(line,0,len);
	
	if (text->nlines > pos) {
		strncpy(line,text->lines[pos],len);
		ret = (int)strlen(line) > len ? len : strlen(line);
		pos++;
		}
	else	ret = -1;

	return(ret);
	}

int TextStream::writeline(char *line)
{	if (((flags&O_ACCMODE) != O_RDWR) && ((flags&O_ACCMODE) != O_WRONLY))
		return(-1);

	text->add_line(line);
	pos = text->nlines;
	return(strlen(line));
	}
