/* ----------------------------------------------------------------------------
 * view.C (c) 92 Michael Kraehe
 */

extern "C" {
#include <stdio.h>
#include <string.h>
#include <memory.h>
#ifndef MINT
#include <malloc.h>
#endif
#include <signal.h>
}
#include "twin.h"

int view_debug_loop = 0;
int view_debug_expose = 0;
int view_debug_init = 0;
int view_debug_map = 0;
int view_debug_event = 0;

View::View(char *InitName,int InitX=0,int InitY=0, int InitB=0,int InitH=0)
{	if (view_debug_init) fprintf(stdout,"Init(%s,%d,%d,%d,%d)\n\r",
		InitName,InitX,InitY,InitB,InitH);

	name = strdup(InitName);
	x = InitX;
	y = InitY;
	b = InitB;
	h = InitH;
	parent = 0;
	visible = 0;
	transaction = 0;
	}

int View::RealX(int vx)
{	if (parent)
		return(parent->RealX(vx+x));
	else	return(vx+x);
	}

int View::RealY(int vy)
{	if (parent)
		return(parent->RealY(vy+y));
	else	return(vy+y);
	}

void View::set_cursor(int cx,int cy)
{	if (visible && parent)
		parent->set_cursor(cx+x,cy+y);
	}

void View::move(int MoveX,int MoveY)
{	int v;

	v = visible;
	if (v) {
		unmap();
		VRoot.expose(RealX(0),RealY(0),b,h);
		}
	x=MoveX;
	y=MoveY;
	if (v) {
		map(parent);
		VRoot.expose(RealX(0),RealY(0),b,h);
		}
	}

void View::add(View *view) { }
void View::del(View *view) { }
void View::up(View *view) { }
void View::down(View *view) { }

void View::map(View *par)
{	if (view_debug_init) fprintf(stdout,"Map(%s,%d,%d,%d,%d)\n\r",
		name,x,y,b,h);

	if (par) parent =par;
	if (!visible) {
		visible = 1;
		VRoot.expose(RealX(0),RealY(0),b,h);
		}
	else	visible = 1;
	}

void View::unmap()
{	if (view_debug_init) fprintf(stdout,"UnMap(%s,%d,%d,%d,%d)\n\r",
		name,x,y,b,h);

	if (visible) {
		visible = 0;
		VRoot.expose(RealX(0),RealY(0),b,h);
		}
	else	visible = 0;
	}

void View::resize(int SizB,int SizH)
{	int vis;
	View *par;

	vis = visible; par = parent;

	if (SizB && SizH) {
		if (vis) unmap();
		b = SizB;
		h = SizH;
		if (vis) map(par);
		}
	}

void View::expose(int ExX,int ExY,int ExB,int ExH)
{	int rx,ry;
	int i;

	if (view_debug_expose) 
		fprintf(stdout,"%sViewExpose %s(%d,%d,%d,%d)\n\r",
			visible ? "" : "No", name,ExX,ExY,ExB,ExH);

	if (!visible) return;

	rx = RealX(ExX);
	ry = RealY(ExY);

	if (!ExB || (ExB+ExX > b)) ExB = b-ExX;
	if (ExH+ExY > h) ExH = h-ExY;

	for (i=0; i<ExH; i++) {
		tw_term_set(rx,ry+i,TW_NORMAL,ExB);
		}
	}

void View::expose()
{	if (!visible) return;

	VRoot.expose(RealX(0),RealY(0),b,h);
	}
	
void View::event(int EvX,int EvY,int EvKey)
{	if (view_debug_event)
                fprintf(stdout,"%sViewEvent %s(%d,%d,%d)\n\r",
                        visible ? "" : "No", name,EvX,EvY,EvKey);
	focus_in(EvKey);
	}

void View::focus_in(int reason)
{	if (VRoot.focus != this) {
		VRoot.focus->focus_out(reason);
		VRoot.focus = this;

		set_cursor(0,0);
		}
	}

void View::focus_out(int reason)
{	}

/* --------------------------------------------------------------------
 * ViewContainer
 */

ViewContainer::ViewContainer(char *InitName,int InitX, int InitY,
		int InitB,int InitH)
	: View(InitName,InitX,InitY,InitB,InitH)
{	viewlist = 0;
	}	

void ViewContainer::add(View *view)
{	struct ViewList **v;
	struct ViewList *vnew;
	
	vnew = (struct ViewList *)setmalloc(sizeof(struct ViewList));

	view->parent = this;
	vnew->view = view;
	vnew->next = 0;

	v = &viewlist;
	while (*v) 
		v = &((*v)->next);
	*v = vnew;

	if (visible && view->visible)
		VRoot.expose(view->RealX(0), view->RealY(0),
			view->b, view->h);
	}

void ViewContainer::del(View *view)
{	struct ViewList **v;
	struct ViewList *vdel;

	v = &viewlist;

	while (*v) {
		if ((*v)->view == view) {
			vdel = *v;
			*v = (*v)->next;
			free(vdel);
			}
		else	v = &((*v)->next);
		}

	if (visible && view->visible)
		VRoot.expose(view->RealX(0), view->RealY(0),
			view->b, view->h);
	}

void ViewContainer::down(View *view)
{	struct ViewList **v;
	struct ViewList *vtmp;

	v = &viewlist;
	vtmp = 0;

	while (*v) {
		if ((*v)->view == view) {
			vtmp = *v;
			*v = (*v)->next;
			}
		else {	v = &((*v)->next);
			}
		}

	vtmp -> next = viewlist;
	viewlist = vtmp;

	if (visible && view->visible)
		VRoot.expose(view->RealX(0), view->RealY(0),
			view->b, view->h);
	}

void ViewContainer::up(View *view)
{	struct ViewList **v;
	struct ViewList *vtmp;

	v = &viewlist;
	vtmp = 0;

	while (*v) {
		if ((*v)->view == view) {
			vtmp = *v;
			*v = (*v)->next;
			}
		else	v = &((*v)->next);
		}

	*v = vtmp;
	vtmp->next = 0;

	if (visible && view->visible)
		VRoot.expose(view->RealX(0), view->RealY(0),
			view->b, view->h);
	}


void ViewContainer::resize(int SizB,int SizH)
{	int vis;	
	View *par;
	struct ViewList *v;

	vis = visible; par = parent;

	if (SizB && SizH) {
		if (vis) unmap();
		b = SizB;
		h = SizH;
		
		v = viewlist;
		while (v) {
			v->view->resize(0,0);
			v = v->next;
			}
		if (vis) map(par);
		}
	}

void ViewContainer::sub_expose(int ExX,int ExY,int ExB,int ExH)
{	int rx,ry;
	int ex,ey,eb,eh;
	struct ViewList *v;

	if (view_debug_expose) 
		fprintf(stdout,"%sContExpose %s(%d,%d,%d,%d)\n\r",
			visible ? "" : "No", name,ExX,ExY,ExB,ExH);

	if (!visible) return;

	rx = RealX(ExX);
	ry = RealY(ExY);

	if (!ExB || (ExB+ExX > b)) ExB = b-ExX;
	if (!ExH || (ExH+ExY > h)) ExH = h-ExY;

	v = viewlist;
	while (v) {
		if (v->view->visible) {
			ex = v->view->x;
			ey = v->view->y;
			eb = v->view->b;
			eh = v->view->h;
			
			if (ex < ExX) {
				eb -= ExX-ex;
				ex = ExX;
				}
			if (ey < ExY) {
				eh -= ExY-ey;
				ey = ExY;
				}
			if (ex+eb > ExX+ExB) eb = (ExX+ExB) - ex;
			if (ey+eh > ExY+ExH) eh = (ExY+ExH) - ey;

			if ((eb > 0) && (eh > 0))
				v->view->expose(ex-v->view->x,
					ey-v->view->y,eb,eh);
			}
		v = v->next;
		}
	}

void ViewContainer::expose(int ExX,int ExY,int ExB,int ExH)
{	View::expose(ExX,ExY,ExB,ExH);
	sub_expose(ExX,ExY,ExB,ExH);
	}

View *ViewContainer::find(int EvX,int EvY)
{	struct ViewList *v;
	struct View *tmp;

	v = viewlist; tmp = 0;
	while (v) {
		if ((v->view->visible) &&
		    (v->view->x <= EvX) &&
		    (v->view->x+v->view->b > EvX) &&
		    (v->view->y <= EvY) &&
		    (v->view->y+v->view->h > EvY)) {
			tmp = v->view;
			}
		v = v->next;
		}

	return(tmp);
	}

void ViewContainer::event(int EvX,int EvY,int EvKey)
{	struct View *tmp;

	if (view_debug_event)
                fprintf(stdout,"%sContEvent %s(%d,%d,%d)\n\r",
                        visible ? "" : "No", name,EvX,EvY,EvKey);

	if ((tmp=find(EvX,EvY)) && (tmp->visible == 1))
		tmp->event(EvX-tmp->x, EvY-tmp->y,EvKey);
	}

void ViewContainer::focus_in(int reason)
{       if (viewlist) {
                viewlist->view->focus_in(reason);
                }
        }

/* --------------------------------------------------------------------
 * ViewBack
 */

ViewBack::ViewBack(char *InitName,int InitX, int InitY, int InitB, int InitH)
	: ViewContainer(InitName,InitX,InitY,InitB,InitH)
{	background = 0;
	free_back = 0;
	}

void ViewBack::back_expose(int ExX,int ExY,int ExB,int ExH)
{       int rx,ry;
        int i,j;
	char **back;
	char *line;
	char *p;

        if (view_debug_expose) 
                fprintf(stdout,"%sMaskExpose %s(%d,%d,%d,%d,%d,%d)\n\r",
                        visible ? "" : "No", name,ExX,ExY,ExB,ExH,b,h);

        if (!visible) return;

        rx = RealX(0);
        ry = RealY(0);

        if (!ExB || (ExB+ExX > b)) ExB = b-ExX;
        if (!ExH || (ExH+ExY > h)) ExH = h-ExY;

	back = background; line = (char*)setmalloc(b+1);

	for (i=0; i<h; i++) {
		memset(line,' ',b); line[b]=0;

		if (back && *back) {
			p=*back;

			for (j=0; j < b; j++) {
				if (!*p) break;

				if (*p == '\t') {
					j = ((j/8)+1)*8;
					}
				else line[j] = *p;

				p++;
				}
			back++;
			}
		if ((i >= ExY) && (i < ExY+ExH)) {
			line[ExX+ExB] = 0;
			tw_term_puts(rx+ExX,ry+i,TW_NORMAL,line+ExX);
        		if (view_debug_expose) 
                		fprintf(stdout,"%s\n\r", line+ExX);
			}
		if (i >= ExY+ExH) break;
		}
	free(line);
	}

void ViewBack::expose(int ExX,int ExY,int ExB,int ExH)
{	back_expose(ExX,ExY,ExB,ExH);
	sub_expose(ExX,ExY,ExB,ExH);
        }


/* --------------------------------------------------------------------
 * ViewRoot
 */

void VR_winch(int type)
{	signal(SIGWINCH,VR_winch);
	tw_term_getsize();
	tw_term_dirty();
	VRoot.resize(max_x,max_y);
	tw_term_refresh();
	}

ViewRoot::ViewRoot(char *Name) : ViewBack(Name,0,0,0,0)
{	tw_term_open();
	b = max_x;
	b = max_y;
	visible = 1;
	focus = &VRoot;
	tw_term_name(Name);
	signal(SIGWINCH,VR_winch);
	}

ViewRoot::~ViewRoot()
{	tw_term_close();
	}

void ViewRoot::set_cursor(int cx,int cy)
{	tw_term_cursor(cx,cy);
	}

void ViewRoot::focus_out(int reason)
{	}

int MausX,MausY,MausB;

int ViewRoot::eventloop()
{	int event;
	static int tx;
	static int ty;

#ifndef SIGWINCH
	tw_term_getsize();
#endif

	if ((b != max_x) || (h != max_y)) {
		tw_term_dirty();
		resize(max_x,max_y);
		}
	/* else	expose(0,0,b,h); */

	if (!visible) map();

	if (view_debug_loop) {
		fflush(stdout);
		event = tw_term_getch();
		tw_term_dirty();
		}
	else
	if (view_debug_expose || view_debug_init 
	 || view_debug_map || view_debug_event) {
		fprintf(stdout,"\nVROOT...\n");
		fflush(stdout);
		}

	tw_term_refresh();
	event = tw_term_getch();

	if (view_debug_event)
                fprintf(stdout,"%sVRootEvent %s(%d,%d,%d)\n\r",
                        visible ? "" : "No", name,curterm->x,curterm->y,event);

	if (event == 10) {
		tw_term_dirty();
		}
	else
	if ((event>=TERM_MAUS1) && (event<=TERM_MAUS3)) {
		tx = curterm->x;
		ty = curterm->y;
		MausB = event;
		}
	else if (event == TERM_REL) {
		MausX = curterm->x;
		MausY = curterm->y;
		VRoot.event(tx,ty,MausB);
		}
	else if (event)
		VRoot.event(curterm->x,curterm->y,event);

	return(event);
	}
